
; Cells.

; The basic unit in topological contructions is the "cell."  A cell
; has a particular dimension; cells of dimension 0 are vertices, of
; dimension 1 are edges, 2 faces, 3 solids, etc.
;
; The boundary of a cell is a set of other cells of lower dimension.
; A vertex has no boundary, an edge's boundary consists of two
; vertices, a face's boundary is a set of edges, etc.
;
; For any cell x, (coboundary x) is the set of cells y for which x is
; on (boundary y).
;
; Each cell comes in two versions, one with positive "polarity" and
; one with negative polarity.  NEGATE takes a cell to its
; corresponding opposite-polarity cell.  The boundary and coboundary
; operators are linear with respect to polarity: if x belongs to
; (boundary y), then (negate x) belongs to (boundary (negate y)).
;
; Every cell C must satisfy the "zero boundary-of-boundary
; constraint": for each cell D in (boundary C), if E is on the
; boundary of D, then (negate E) is on the boundary of exactly one
; other cell in (boundary C).  A dual constraint holds for
; coboundaries.

(define (make-cell bs)
  (let ((dim (dimension (car bs))))
    (for-each (lambda (b)
                (if (not (= (dimension b) dim))
                    (error "dimension mismatch" bs)))
              (cdr bs))
    (if (= dim 0)	;Making an edge out of two vertices
	(if (not (and (= (length bs) 2)
		      (= (+ (sign (car bs)) (sign (cadr bs))) 0)))
	    (error "edge trouble - make-cell" (bb bs)))
	(let ((uids (bb bs)))
	  ;; Approximate check.
	  (if (not (= (reduce + 0 (reduce append '() uids)) 0))
	      (error "boundary of boundary is nonzero - make-cell" uids))))
    (really-make-cell bs
		      (h/ (reduce (lambda (b sum) (h+ (centroid b) sum))
				  (vector 0 0 0 1)
				  bs)
			  (length bs)))))

(define (make-vertex coords)
  (if (not (vector? coords))
      (error "invalid coordinates" coords))
  (really-make-cell '() coords))

; Cell data structure

(define-record-type cell type/cell
  (create-cell bs centroid dim cs memos uid)
  cell?
  (bs boundary)
  (centroid centroid)
  (dim dimension)
  (cs coboundary set-coboundary!)
  (memos cell-memos set-cell-memos!)
  (uid unique-id)
  (neg negate set-negative!))

(define-record-discloser type/cell
  (lambda (cell)
    (let ((uid (unique-id cell))
	  (bs (map unique-id (boundary cell))))
      (case (dimension cell)
	((0) (list 'vertex uid (centroid cell)))
	((1) (list 'edge uid bs))
	((2) (list 'face uid bs))
	((3) (list 'solid uid bs))
	(else (list 'cell (dimension cell) uid bs))))))

(define (really-make-cell bs+ centroid)
  (let ((dim (if (null? bs+) 0 (+ 1 (dimension (car bs+)))))
	(uid (new-unique-id))
	(bs- (map negate (reverse bs+))))
    (let ((cell+ (create-cell bs+ centroid dim '() '() uid))
	  (cell- (create-cell bs- centroid dim '() '() (- 0 uid))))
      (set-negative! cell+ cell-)
      (set-negative! cell- cell+)
      (for-each (lambda (b)
		  (add-cobounder! b cell+))
		bs+)
      (for-each (lambda (b)
		  (add-cobounder! b cell-))
		bs-)
      (if *debug-make-cell?*
	  (format #t "(make-cell ~:s) => ~s~%"
		  (map unique-id bs+)
		  uid))
      cell+)))

(define *debug-make-cell?* #f)

(define (add-cobounder! cell cobounder)
  (set-coboundary! cell (cons cobounder (coboundary cell))))

(define (sign cell)
  (if (> (unique-id cell) 0) 1 -1))

(define new-unique-id
  (let ((id 0))
    (lambda ()
      (set! id (+ id 1))
      id)))

(define (cell-memo cell proc neg)
  (let* ((memos (cell-memos cell))
	 (probe (assq proc memos)))
    (if probe
	(cdr probe)
	(let ((value (proc cell)))
	  (set-cell-memos! cell
			   (cons (cons proc value) memos))
	  (set-cell-memos! (negate cell)
			   (cons (cons proc (neg value))
				 (cell-memos (negate cell))))
	  value))))


(define (coordinates vertex)
  (if (vertex? vertex)
      (centroid vertex)
      (error "invalid vertex" vertex)))

(define (vertex? cell) (= (dimension cell) 0))
(define (edge? cell)   (= (dimension cell) 1))
(define (face? cell)   (= (dimension cell) 2))
(define (solid? cell)  (= (dimension cell) 3))

(define (memoize-at-cell proc neg)
  (lambda (cell)
    (cell-memo cell proc neg)))

; Debugging routine: returns list of lists of uid's for boundary's boundary.

(define (bb cell-list)
  (map (lambda (c) (map unique-id (boundary c)))
       cell-list))


; (boundary (make-edge tail head)) = head - tail
; The coboundary of a vertex is the set of edges whose head is that vertex.

(define (make-edge tail head)
  (if (and (vertex? tail)
	   (vertex? head))
      (make-cell (list head (negate tail)))
      (error "invalid arguments to make-edge"
	     (dimension tail) (dimension head))))

; An edge's head is the vertex on its boundary with positive polarity.

(define (edge-head edge)
  (let ((bs (boundary edge)))
    (if (= (sign (car bs)) 1)
	(car bs)
	(cadr bs))))

; An edge's tail is the negation of the vertex on its boundary with
; negative polarity.

(define (edge-tail edge)
  (negate (let ((bs (boundary edge)))
	    (if (= (sign (car bs)) 1)
		(cadr bs)
		(car bs)))))

(define (find-edge tail head)
  (let loop ((es (coboundary head)))
    (cond ((null? es)
	   (make-edge tail head))
	  ((eq? tail (edge-tail (car es)))
	   (car es))
	  (else (loop (cdr es))))))

(define (edge-length cell)
  (if (edge? cell)
      (let ((b (boundary cell)))
	(hnorm (h- (coordinates (car b)) (coordinates (cadr b)))))
      (edge-length (car (boundary cell)))))

; A complex is a set of cells.

(define (complex-cells complex)
  (cond ((complex-not-cell? complex)
	 complex)
	(else (list complex))))

(define (complex-not-cell? complex)
  (or (null? complex) (pair? complex)))

; Extend a mapping on cells to a mapping on complexes.

(define (extend-to-complexes f)
  (lambda (complex)
    (if (complex-not-cell? complex)
	(map f complex)
	(f complex))))

; Mapping procedures.

(define (for-each-cell proc test complex)
  (letrec ((recur (memoize (lambda (cell)
                             (if (test cell)
                                 (proc cell)
                                 (for-each recur (boundary cell)))
                             #t)
			   (lambda (cell) #t)))) ;negate
    (for-each recur (complex-cells complex))
    (values)))
                
(define (map-over-cells combiner initial complex dim)
  (map-over-cells-1 combiner
                    initial
                    (lambda (cell) (= (dimension cell) dim))
                    complex))

(define (map-over-cells-1 combiner initial test complex)
  (let ((result initial))
    (for-each-cell (lambda (cell)
                     (set! result (combiner cell result))
                     #t)
                   test
                   complex)
    result))

(define (all-cells complex dim)
  (map-over-cells cons '() complex dim))

(define (all-vertices complex) (all-cells complex 0))
(define (all-edges complex) (all-cells complex 1))
(define (all-faces complex) (all-cells complex 2))

(define (for-each-edge proc complex)
  (for-each-cell proc edge? complex))

(define (for-each-face proc complex)
  (for-each-cell proc face? complex))

(define (for-each-solid proc complex)
  (for-each-cell proc solid? complex))


; Homomorphically extend a mapping on vertices to a mapping on complexes.

(define (induced-mapping proc)
  (letrec ((recur
            (memoize (lambda (cell)
                       (if (vertex? cell)
			   (proc cell)
                           (make-cell (map recur (boundary cell)))))
		     negate)))
    (extend-to-complexes recur)))


; All of a face's vertices, in order.

(define (face-vertices face)
  (map edge-head (ordered-edges face)))

; Make a face from a set of vertices.  (Sort of an inverse to face-vertices.)

(define (face-from-vertices vs)
  (make-cell (map-cyclically find-edge vs)))


; Permute a face's edges so that they form a cycle.  If e2 immediately
; follows e1 in the list, then e1's head is e2's tail.

(define ordered-edges
  (memoize-at-cell (lambda (face)
		     (form-cycle (boundary face)
				 (lambda (edge1 edge2)
				   (eq? (edge-head edge1)
					(edge-tail edge2)))))
		   (lambda (es)
		     (reverse (map negate es)))))

; Permute the edges incident with a vertex so that adjacent edges
; share a face.

(define (co-ordered-edges vertex)
  (form-cycle (coboundary vertex)
	      (lambda (edge1 edge2)
		(intersect? (coboundary (negate edge1))
			    (coboundary edge2)))))

(define (form-cycle es adjacent?)
  (let loop ((es (cdr es)) (z (list (car es))))
    (if (null? es)
	(reverse z)
	(let ((probe (find-if (lambda (edge)
				(adjacent? (car z) edge))
			      es)))
	  (if probe
	      (loop (remq probe es)
		    (cons probe z))
	      (error "cycle lossage"
		     (bb es) (bb z)))))))

; Given a set of cells, INVERT-AS-NECESSARY returns the same set with
; the polarities of some of the cells reversed so as to enforce the
; zero boundary-of-boundary constraint (see above).

(define (invert-as-necessary cells)
  (if (vertex? (car cells))
      (let ((v1 (car cells))
	    (v2 (cadr cells)))
	(if (not (= (length cells) 2))
	    (error "edge lossage - invert-as-necessary" (map unique-id cells)))
	(list v1
	      (if (= (sign v1) (sign v2)) (negate v2) v2)))
      (invert-as-necessary1 (cdr cells) (list (car cells)))))

; (INVERT-AS-NECESSARY1 x y) is like (INVERT-AS-NECESSARY (append x y)
; except that the polarity of cells in y is unchanged.

(define (invert-as-necessary1 cells z)
  (if (null? cells)
      z
      ;; Find a cell that shares a boundary with a cell that's already
      ;; been seen.
      (let ((probe (find-if (lambda (c)
			      (find-cancellation c z))
			    cells)))
	(if probe
	    (invert-as-necessary1 (remq probe cells) (cons probe z))
	    (let ((probe (find-if (lambda (c)
				    (find-cancellation (negate c) z))
				  cells)))
	      (if probe
		  (invert-as-necessary1 (remq probe cells)
					(cons (negate probe) z))
		  (error "lossage in invert-as-necessary1"
			 (bb cells) (bb z))))))))

(define (find-cancellation cell cells)
  (let ((b (boundary (negate cell))))
    (find-if (lambda (cell)
	       (intersect? b (boundary cell)))
	     cells)))

; --------------------
; Memoization.

(define (memoize f negate-result)
  (let ((memos '()))
    (lambda (x)
      (let* ((pos? (> (sign x) 0))
	     (x+ (if pos? x (negate x)))
	     (probe (assq x+ memos))
	     (results
	      (if probe
		  (cdr probe)
		  (let* ((result (f x+))
			 (results (cons result (negate-result result))))
		    (set! memos
			  (cons (cons x+ results) memos))
		    results))))
	(if pos?
	    (car results)
	    (cdr results))))))


; --------------------
; Centroids and normals.

; Centroid (center of mass)

; This is no longer needed because the centroid is now computed when the
; cell is created.
;(define (compute-centroid cell)
;  (let ((final-state
;         (map-over-cells (lambda (vertex state)
;                           ;; state is (sum . count)
;                           (cons (h+ (car state) (coordinates vertex))
;                                 (+ (cdr state) 1)))
;                         (cons (vector 0 0 0 1) 0)
;                         cell
;                         0)))
;    (h/ (car final-state) (cdr final-state))))
;
;(define centroid (memoize-at-cell compute-centroid (lambda (c) c)))

; Arrange for normal to point away from the center of a solid.

(define (compute-face-normal face)
  (really-compute-face-normal face))

(define face-normal
  (memoize-at-cell compute-face-normal (lambda (n) (h* n -1))))

; The normal as computed here will point towards a viewer who sees the
; edges going around counterclockwise.

(define (really-compute-face-normal face) ;Bad algorithm.  Should use Newell's.
  (let* ((vs (face-vertices face))
	 (v1 (car vs))
	 (v2 (cadr vs))
	 (v3 (caddr vs)))
    (hcross (h- (coordinates v2) (coordinates v3))
            (h- (coordinates v2) (coordinates v1)))))

; Negate a solid, if necessary, in order to make its face normals
; point out of the solid.

(define (reorient-if-necessary solid)
  (if (let ((face (car (boundary solid))))
	(< (hdot (h- (centroid face) (centroid solid)) (face-normal face))
	   0))
      (negate solid)
      solid))


; Code from EIGHT
;(define (plane-equation list-of-coordinates)
;  (do ((chain list-of-coordinates (cdr chain))
;       (a 0) (b 0) (c 0)
;       (first (car list-of-coordinates)))
;      ((null? (cdr chain))
;       (vector a b c
;               (- (+ (* a (h-x first))
;                     (* b (h-y first))
;                     (* c (h-z first))))))
;    (let ((i (car chain))
;          (j (cadr chain)))
;      (setq a (+ a (* (- (h-y i) (h-y j))
;                      (+ (h-z i) (h-z j)))))
;      (setq b (+ b (* (- (h-z i) (h-z j))
;                      (+ (h-x i) (h-x j)))))
;      (setq c (+ c (* (- (h-x i) (h-x j))
;                      (+ (h-y i) (h-y j))))))))

; --------------------
; Output utilities

; Hidden surfaces... this doesn't work so well for non-convex solids...

(define (for-each-visible-face proc complex view)
  (for-each-solid (lambda (solid)
		    (for-each (lambda (face)
				(cond ((visible? face view)
				       (proc face))))
			      (boundary (reorient-if-necessary solid))))
		  complex))


; Very low-tech hidden surface predicate.

(define (visible? face view)
  (> (hdot (h- (view-position view) (centroid face)) (face-normal face))
     0))

; --------------------
; Debugging routine.

(define (show cell)
  (set! *cells* '())
  (let ((dim (dimension cell)))
    (letrec ((recur (lambda (cell)
		      (let ((indent (* 2 (- dim (dimension cell)))))
			(format #t "~&~vt(~d = "
				indent
				(unique-id cell))
			(if (vertex? cell)
			    (format #t "[~{~6,1f~^~}]"
				    (vector->list (coordinates cell)))
			    (format #t "~s"
				    (map unique-id (boundary cell))))
			(for-each recur
				  (boundary cell))
			(format #t ")")
			(if (not (cel (unique-id cell)))
			    (set! *cells* (cons (cons (unique-id cell) cell)
						*cells*)))
			))))
      (recur cell)
      (format #t "~%")
      (values))))
(define *cells* '())
(define (cel uid)
  (let ((probe (assv uid *cells*)))
    (if probe (cdr probe) #f)))

;(define (show cell)
;  (let ((dim (dimension cell)))
;    (letrec ((recur (memoize
;                     (lambda (cell)
;                       (let ((indent (* 2 (- dim (dimension cell)))))
;                         (format #t "~&~vt(~d = "
;                                 indent
;                                 (unique-id cell))
;                         (if (vertex? cell)
;                             (format #t "[~{~6,1f~^~}]"
;                                     (vector->list (coordinates cell)))
;                             (format #t "~s"
;                                     (map unique-id (boundary cell))))
;                         (for-each recur
;                                   (boundary cell))
;                         (format #t ")")))
;                     (lambda (cell)
;                       #t))))
;      (recur cell)
;      (format #t "~%")
;      (values))))

(define (sho name x) (format #t "~&~a: ~s~%" name x) x)
