
; Some general-purpose constructions.

; Product of two complexes.

(define (product complex1 complex2 vector*)
  (letrec ((recur
	    (memoize
	     (lambda (cell1)
	       (memoize
		(lambda (cell2)
		  (if (and (vertex? cell1)
			   (vertex? cell2))
		      (make-vertex (vector* (coordinates cell1)
					    (coordinates cell2)))
		      ;; d(A x B) = (A x dB) + (dA x B)
		      (make-cell (invert-as-necessary
				  (append (map (lambda (b)
						 ((recur cell1) b))
					       (boundary cell2))
					  (map (lambda (a)
						 ((recur a) cell2))
					       (boundary cell1)))))))
		negate))
	     (lambda (f)
	       (lambda (cell) (negate (f cell)))))))
    (let ((cells
	   (apply append
		  (map (lambda (cell1)
			 (let ((*cell1 (recur cell1)))
			   (map (lambda (cell2)
				  (*cell1 cell2))
				(complex-cells complex2))))
		       (complex-cells complex1)))))
      (if (and (not (null? cells))
	       (null? (cdr cells)))
	  (car cells)
	  cells))))

(define (cone cell apex)
  (let ((a (list (negate apex))))
    (letrec ((recur
	      (memoize (lambda (cell)
			 (make-cell
			  (cons cell
				(if (vertex? cell)
				    a
				    (map recur (boundary (negate cell)))))))
		       negate)))
      (recur cell))))

; proc should map vertices to vertices.  Edges are established between
; corresponding old and new vertices, faces between corresponding
; edges, etc.
; [Actually, this could be done easily using PRODUCT.]

(define (extrude cell proc)
  (let ((proc (induced-mapping proc)))
    (letrec ((recur
	      (memoize
	       (lambda (cell)
		 ;; append doesn't work here; must invert some cells.
		 (make-cell (invert-as-necessary1
			         (map recur (boundary cell))
				 (list cell
				       (negate (proc cell))))))
	       negate)))
      (recur cell))))

; Form the dual of a cell of arbitrary dimension.

(define (dual p)
  (let ((dim (- (dimension p) 1)))
    (letrec ((recur
	      (memoize
	       (lambda (cell)
		 (if (= (dimension cell) dim) 
		     (make-vertex (centroid cell))
		     (make-cell (invert-as-necessary
				     (map recur (coboundary cell))))))
	       negate)))
      (make-cell (invert-as-necessary (map recur (all-vertices p)))))))

; FULL-TRUNCATE truncates a solid by connecting the midpoints of its
; edges.  E.g. it takes a cube or octahedron to a cuboctahedron.

(define (full-truncate cell)
  ;; vertex -> face with one edge per vertex's cobounder
  ;; face -> face with same # edges
  ;; edge -> 2 edges
  (let ((edge->midpoint
         (memoize (lambda (edge)
                    (make-vertex (centroid edge)))
		  (lambda (v) v))))
    ;; Solid
    (make-cell
     (invert-as-necessary1 (map (lambda (vertex)
				  (face-from-vertices
				   (map edge->midpoint
					(co-ordered-edges vertex))))
				(all-vertices cell))
			   (map (lambda (face)
				  (face-from-vertices
				   (map edge->midpoint (ordered-edges face))))
				(boundary cell))))))

; Generalized truncation: make edges shorter by a specified ratio.
; ratio = 1 means no operation.
; ratio = 0 means full-truncate.

(define (general-truncate solid ratio)
  (let ((shorten (memoize (lambda (edge)
			    (dilate edge ratio))
			  negate)))
    (make-cell
     (invert-as-necessary1
       (map (lambda (vertex)
	      (face-from-vertices
	       (map (lambda (edge)
		      (edge-head (shorten edge)))
		    (co-ordered-edges vertex))))
	    (all-vertices solid))
       (map (lambda (face)
	      (let ((es (map shorten (ordered-edges face))))
		(make-cell (append es
				   (map-cyclically (lambda (e1 e2)
						     (find-edge
						      (edge-head e1)
						      (edge-tail e2)))
						   es)))))
	    (boundary solid))))))

; Make a copy of cell, scaled by given scale factor, with coincident
; centroid.

(define (dilate cell scale)
  (let ((c (centroid cell)))
    (copy cell
	  (mm* (translation c)
	       (dilatation scale)
	       (translation (h* c -1))))))

(define (shared-bounder cell1 cell2)
  (unique-intersection (boundary cell1) (boundary cell2)))

(define (unique-intersection l1 l2)
  (let ((l (intersectq l1 l2)))
    (cond ((null? l) (error "no intersection" l1 l2))
          ((null? (cdr l)) (car l))
          (else (error "intersection not unique" l1 l2)))))


; Standard truncation, attempting to make all edges the same length.
; This produces Archimedean solids from Platonic solids.
; It doesn't always work (e.g. great rhombicosidodecahedron).

(define (semi-truncate cell)
  ;; vertex -> new face
  ;; edge -> shorter edge
  ;; face -> smaller face with twice as many edges
  (general-truncate cell (semi-truncation-ratio cell)))

; Compute d/e where
;   e = edge length,
;   d = (e - d) sin alpha, and
;   alpha = angle between adjacent edges f a face.

(define (semi-truncation-ratio cell)
  (if (face? cell)
      (let* ((vs (face-vertices cell))
	     (a (coordinates (car vs)))
	     (o (coordinates (cadr vs)))
	     (b (coordinates (caddr vs)))
	     (alpha (/ (angle-between (h- a o) (h- b o)) 2))
	     (sin-alpha (sin alpha)))
	(/ sin-alpha (+ 1 sin-alpha)))
      (semi-truncation-ratio (car (boundary cell)))))


; --------------------
; Some particular shapes.

; n-gon generates a regular polygon in the x-y plane.

(define (n-gon n r)
  (let ((dtheta (/ (* 2 pi) n)))
    (do ((i 0 (+ i 1))
         (theta 0.0 (+ theta dtheta))
         (vs '() (cons (make-vertex (vector (* r (cos theta))
                                            (* r (sin theta))
                                            0
                                            1))
                       vs)))
        ((>= i n)
	 (make-cell (map-cyclically make-edge (reverse vs)))))))

(define (prism cell height dir)
  (let ((v (h* (unitize dir) height)))
    (extrude (copy cell (translation (h/ v 2)))
	     (matrix->procedure (translation (h* v -1))))))

(define (copy complex m)
  ((induced-mapping (matrix->procedure m)) complex))

(define (matrix->procedure m)
  (lambda (v)
    (make-vertex (mv* m (coordinates v)))))

; Regular prism - sides square.

(define (drum n e)
  (let ((g (n-gon n (/ e (* 2 (sin (/ pi n)))))))
    (prism g e (face-normal g))))


; Platonic solids

(define (make-tetrahedron e)            ;centered at origin
  (let* ((face-radius (/ e (sqrt 3)))
         (radius (/ e (* 2 (sqrt (/ 2 3)))))
         (base (n-gon 3 face-radius)))
    (cone (copy base (translation (vector 0 0 (- radius e) 1)))
          (make-vertex (vector 0 0 radius 1)))))

(define (make-cube e) (drum 4 e))


; Make an icosahedron (20 triangular vertices).  There has got to be
; an easier way to do this, but I can't think of one.
;
; The general strategy is as follows:
;
; 1. Form three mutually perpendicular golden rectangles.  The 12
; vertices of these rectangles form the vertices of the icosahedron.
; 
; 2. Form all triangular faces that run from the short edge from one
; golden rectangle to a vertex of another.  There are 6 such edges,
; with two faces incident with each, giving a total of 12 faces.
;
; 3. Form all of the triangular faces that have one vertex from each
; of the three golden rectangles.  There are eight such faces, and
; they lie in a cubic configuration (i.e. their centroids form the
; vertices of a cube).

(define (make-icosahedron e)
  (let ((stuff (let ((e1 (let ((v1 (make-vertex (vector (/ e 2)
							(* e (+ 1.0 (sqrt 5.0))
							   1/4)
							0
							1))))
			   (make-edge v1 (copy v1 yz-mirror))))
		     (m (mm* (x-rotation (/ pi 2))
			     (y-rotation (/ pi 2)))))
		 (do ((e1 e1 (copy e1 m))
		      (e2 (copy e1 (z-rotation pi))
			  (copy e2 m))
		      (l '() (cons (cons e1 e2) l))
		      (i 0 (+ i 1)))
		     ((= i 3) l)))))
    (make-cell
     (append (reduce append '()
		     (map-cyclically (lambda (es1 es2)
				       (list (cone (negate (car es2))
						   (edge-head (cdr es1)))
					     (cone (cdr es2)
						   (edge-head (car es1)))
					     (cone (car es2)
						   (edge-tail (car es1)))
					     (cone (negate (cdr es2))
						   (edge-tail (cdr es1)))))
				     stuff))
	     (map (let ((s1 (car stuff))
			(s2 (cadr stuff))
			(s3 (caddr stuff)))
		    (lambda (y)
		      (apply (lambda (v1 e1 v2 e2 v3 e3 n)
			       (n (face-from-vertices
				   (list (v1 (e1 s1))
					 (v2 (e2 s2))
					 (v3 (e3 s3))))))
			     y)))
		  (let ((w (lambda (v)
			     (if (eq? v edge-head) edge-tail edge-head)))
			(f (lambda (e)
			     (if (eq? e car) cdr car)))
			(m (lambda (n)
			     (if (eq? n negate) (lambda (x) x) negate)))
			(a (lambda (f y)
			     (append y (map (lambda (z) (apply f z)) y)))))
		    (a (lambda (v1 e1 v2 e2 v3 e3 n) ;Three cubic symmetries.
			 (list v1 e1 (w v2) e2 (w v3) (f e3) (m n)))
		       (a (lambda (v1 e1 v2 e2 v3 e3 n)
			    (list (w v1) e1 (w v2) (f e2) v3 e3 (m n)))
			  (a (lambda (v1 e1 v2 e2 v3 e3 n)
			       (list v1 (f e1) v2 (f e2) v3 (f e3) (m n)))
			     (list (list edge-tail car
					 edge-tail car
					 edge-tail car
					 negate)))))))))))

; (put 'with-color 'scheme-indent-hook 1)
