
; Display routines.  These mediate between the topological data
; structures and 3-D projection routines, on the one hand, and the
; low-level 2-D device output, on the other.

; FRAME makes a wire frame drawing of a cell complex.

(define (frame complex view)
  (clear)
  (frame1 complex view)
  (showit))

(define (frame1 complex view)
  (for-each-edge (lambda (edge)
                   (draw-edge edge view))
                 complex))

;(define (frame1 complex view)
;  (for-each (lambda (cell)
;              (case (dimension cell)
;                ((0) (draw-vertex cell view))
;                ((1) (draw-edge cell view))
;                (else (for-each-face (lambda (face)
;                                       (draw-face face view #f))
;                                     cell))))
;            (complex-cells complex)))

; HIDE is like FRAME, but only shows visible faces.

(define (hide solid view)
  (clear)
  (hide1 solid view)
  (showit))

(define (hide1 solid view)
  (for-each-visible-face (lambda (face)
                           (frame1 face view))
                         solid
                         view))

; A somewhat more clever renderer.
; Draw faces in order from back to front.

(define (hide* complex view)
  (let* ((pos (view-position view))
         (faces (sort-list (all-faces complex)
			   (lambda (f1 f2)
			     (< (hdot (centroid f1) pos)
				(hdot (centroid f2) pos))))))
    (clear)
    (for-each (lambda (face)
                (draw-shaded-face face view))
              faces)
    (showit)))

; SLICK is like HIDE, but colors in all the faces as well.
; A face's color is a function of the number of its edges.

(define (slick complex view)
  (clear)
  (slick1 complex view)
  (showit))

(define (slick1 complex view)
  (for-each-visible-face (lambda (face)
			   (draw-colored-face face (face-color face) view))
			 complex
			 view))

(define (draw-colored-face face color view)
  (with-color color
    (lambda ()
      (draw-face face view #t)))
  (with-color color/black
    (lambda ()
      (frame1 face view))))

(define color/black (allocate-color 0 0 0))
(define color/red   (allocate-color 1 0 0))
(define color/green (allocate-color 0 1 0))
(define color/blue  (allocate-color 0 0 1))
(define color/white (allocate-color 1 1 1))

(define (face-color face)
  (list-ref colors
	    (modulo (- (length (boundary face)) 3)
		    number-of-colors)))

(define colors
  ;; Mumble.  This only works with MTV (not postscript)
  '(:yellow :orange :blue :red :green :purple :brown :pink))

(define number-of-colors (length colors))


; SHADE displays a polyhedron with each face drawn a shade of gray
; determined from its orientation relative to a "sun direction".  Faces
; whose normals point towards the sun show up light in color, while
; faces aimed away from the sun are darker.  (Simple computation from
; dot product of face normal with direction of light source.)

(define (shade complex view)
  (clear)
  (shade1 complex view)
  (showit))

(define (shade1 complex view)
  (for-each-visible-face (lambda (face)
			   (draw-shaded-face face view))
			 complex
			 view))


(define (draw-shaded-face face view)
  (with-gray (* (+ (hdot (unitize (face-normal face))
			 sun-direction)
		   1)
		9/20)
    (lambda ()
      (draw-face face view #t))))

(define sun-direction (unitize (vector 2 -1 3 1)))

; Put some text starting at cell's centroid.

(define (annotate cell string view)
  (let ((p (apply-view view (centroid cell))))
    (scrawl (car p)
	    (cadr p)
	    string)))

; Draw an edge somehow.

(define (draw-edge edge view)
  (let ((bs (boundary edge)))
    (let ((p1 (apply-view view (coordinates (car bs))))
          (p2 (apply-view view (coordinates (cadr bs)))))
      (draw-segment (car p1) (cadr p1) (car p2) (cadr p2)))))

; Fill in a single face.

(define (draw-face face view fill?)
  (draw-polygon
   (map (lambda (vertex)
          (apply-view view (coordinates vertex)))
        (face-vertices face))
   fill?))

; Draw a little circle around a vertex.

(define (draw-vertex vertex view)
  (let ((p (apply-view view (coordinates vertex))))
    (draw-point (car p) (cadr p))))
