; Copyright 1993 Jonathan Rees
;
; Permission to copy this software, in whole or in part, to use this
; software for any lawful purpose, and to redistribute this software
; is granted subject to the restriction that all copies made of this
; software must include this copyright notice in full.
;
; I also request that you send me a copy of any improvements that you
; make to this software so that they may be incorporated within it to
; the benefit of the Scheme community.


; A cheap imitation of Bruce Donald's EIGHT geometric modeling package.

; In Pseudoscheme, do (benchmark-mode) before compiling.

; From: brd   Date: Wed, 4 Nov 92 22:34:50 -0500
; A working hidden surface routine is in
;
; ~brd/b/eight/84-hidden-exper-3.lisp
; [ ~jar/eight/84-hidden-exper-3.lisp ]
;
; and is pretty generic. One caveat though, is that at the time I was
; experimenting with reader macros, so the ! is a special form, i.e.,
;
; !face.normal.x ==> (x (normal face))


;;;; Geometry

(define pi (atan 0.0 -1.0))

;;; Vectors

(define (generate-vector len proc)
  (let ((v (make-vector len)))
    (do ((i 0 (+ i 1)))
        ((>= i len) v)
      (vector-set! v i (proc i)))))

(define (map-vector proc v)
  (generate-vector (vector-length v)
                   (lambda (i) (proc (vector-ref v i)))))

(define (dot v1 v2)
  (let ((len (vector-length v1)))
    (let loop ((sum 0) (i 0))
      (if (< i len)
          (loop (+ sum (careful* (vector-ref v1 i)
                                 (vector-ref v2 i)))
                (+ i 1))
          sum))))

; Multiply two numbers, with the result of multiplication by an exact
; zero being an exact zero.

(define (careful* n1 n2)
  (cond ((= n1 0) 0)
        ((= n2 0) 0)
        (else (* n1 n2))))

;;; Matrices

; A matrix is represented as a vector (column) of rows.

(define matrix-row vector-ref)
(define matrix-nrows vector-length)
(define matrix vector)

(define (matrix-ncolumns m)
  (vector-length (matrix-row m 0)))

(define (matrix-ref matrix row col)
  (vector-ref (matrix-row matrix row) col))

(define (matrix-column m i)
  (generate-vector (matrix-nrows m)
                   (lambda (j)
                     (matrix-ref m j i))))

(define (generate-matrix nrows ncols proc)
  (generate-vector nrows
                   (lambda (i)
                     (generate-vector ncols
                                      (lambda (j)
                                        (proc i j))))))

(define (transpose matrix)
  (generate-matrix (matrix-ncolumns matrix)
                   (matrix-nrows matrix)
    (lambda (i j)
      (matrix-ref matrix j i))))

(define (mv* m v)			;Matrix times column vector
  (generate-vector (vector-length v)
                   (lambda (i)
                     (dot (matrix-row m i) v))))

(define (mm* first . rest)		;Matrix times matrix
  (let loop ((first first) (rest rest))
    (if (null? rest)
        first
        (binary-mm* first (loop (car rest) (cdr rest))))))

(define (binary-mm* m1 m2)
  (generate-matrix (matrix-nrows m1)
                   (matrix-ncolumns m2)
                   (lambda (i j)
                     (dot (matrix-row m1 i) (matrix-column m2 j)))))

(define (identity rank)
  (generate-matrix rank rank
                   (lambda (i j) (if (= i j) 1 0))))

; Debugging routine: display a matrix.

;(define (mm m)
;  (format #t "~{[~{~8,2f~}]~%~}"    ;yippee!  we're common lisp cowboys
;          (map vector->list (vector->list m))))


; Operations on homogeneously represented 3-space vectors.

(define (x-component v) (vector-ref v 0))
(define (y-component v) (vector-ref v 1))
(define (z-component v) (vector-ref v 2))
(define (h-component v) (vector-ref v 3))

(define (h-x v) (/ (x-component v) (h-component v)))
(define (h-y v) (/ (y-component v) (h-component v)))
(define (h-z v) (/ (z-component v) (h-component v)))

(define (h+ v1 v2)                      ;add homogeneous vectors
  (let ((h1 (h-component v1))
        (h2 (h-component v2)))
;    (vector (+ (* (x-component v1) h2)    ;This is the preferred formula,
;               (* (x-component v2) h1))   ;but it has the problem that the
;            (+ (* (y-component v1) h2)    ;h's blow up rapidly...
;               (* (y-component v2) h1))
;            (+ (* (z-component v1) h2)
;               (* (z-component v2) h1))
;            (* h1 h2))
    (vector (+ (/ (x-component v1) h1)
               (/ (x-component v2) h2))
            (+ (/ (y-component v1) h1)
               (/ (y-component v2) h2))
            (+ (/ (z-component v1) h1)
               (/ (z-component v2) h2))
            1)))

(define (h- v1 v2)                      ;subtract homogeneous vectors
  (let ((h1 (h-component v1))
        (h2 (h-component v2)))
    (vector (- (* (x-component v1) h2)
               (* (x-component v2) h1))
            (- (* (y-component v1) h2)
               (* (y-component v2) h1))
            (- (* (z-component v1) h2)
               (* (z-component v2) h1))
            (* h1 h2))))

(define (hdot v1 v2)                    ;homogeneous dot product
  (/ (+ (* (x-component v1) (x-component v2))
        (* (y-component v1) (y-component v2))
        (* (z-component v1) (z-component v2)))
     (* (h-component v1) (h-component v2))))

(define (hnorm v)
  (/ (sqrt (+ (square (x-component v))
              (square (y-component v))
              (square (z-component v))))
     (h-component v)))

(define (square x) (* x x))

(define (unitize v)			;unit vector parallel to v
  (h/ v (hnorm v)))

(define (angle-between v1 v2)
  ;; (acos (/ (hdot v1 v2) (* (hnorm v1) (hnorm v2))))
  (let ((x1 (x-component v1))
	(y1 (y-component v1))
	(z1 (z-component v1))
	(x2 (x-component v2))
	(y2 (y-component v2))
	(z2 (z-component v2)))
    (acos (/ (+ (* x1 x2) (* y1 y2) (* z1 z2))
	     (sqrt (* (+ (* x1 x1) (* y1 y1) (* z1 z1))
		      (+ (* x2 x2) (* y2 y2) (* z2 z2))))))))

(define (hcross v1 v2)                  ;cross product
  (let ((det (lambda (i j)
               (- (* (vector-ref v1 i) (vector-ref v2 j))
                  (* (vector-ref v1 j) (vector-ref v2 i))))))
    (vector (det 1 2)
            (det 2 0)
            (det 0 1)
            (* (h-component v1)
               (h-component v2)))))

(define (h* v n)                        ; scale up
  (vector (* (x-component v) n)
	  (* (y-component v) n)
	  (* (z-component v) n)
	  (h-component v)))

(define (h/ v n)                        ; scale down
  (let ((x (x-component v))
        (y (y-component v))
        (z (z-component v))
        (h (h-component v)))
    (if (or (= n 0)
            (and (exact? x) (exact? y) (exact? z)))
        (vector x y z (* h n))
        (vector (/ x n) (/ y n) (/ z n) h))))

(define (homogenize v)
  (vector (x-component v) (y-component v) (z-component v) 1))

(define (dehomogenize v)
  (let ((h (h-component v)))
    (vector (/ (x-component v) h)
            (/ (y-component v) h)
            (/ (z-component v) h))))

; Matrices for operating on homogeneous coordinates.

(define (translation v)
  (let ((h (h-component v)))
    (matrix (vector h 0 0 (x-component v))
	    (vector 0 h 0 (y-component v))
	    (vector 0 0 h (z-component v))
	    (vector 0 0 0 h))))

(define (z-rotation theta)		;rotation about z axis
  (let ((c (cos theta))
        (s (sin theta)))
    (matrix (vector c (- s) 0 0)
            (vector s    c  0 0)
            (vector 0    0  1 0)
            (vector 0    0  0 1))))

(define (x-rotation theta)		;rotation about x axis
  (let ((c (cos theta))
        (s (sin theta)))
    (matrix (vector 1 0    0  0)
            (vector 0 c (- s) 0)
            (vector 0 s    c  0)
            (vector 0 0    0  1))))

(define (y-rotation theta)		;rotation about y axis
  (let ((c (cos theta))
        (s (sin theta)))
    (matrix (vector    c  0 s 0)
            (vector    0  1 0 0)
            (vector (- s) 0 c 0)
            (vector    0  0 0 1))))

(define (mirror axis)
  (generate-matrix 4 4 (lambda (i j)
			 (if (= i j) (if (= i axis) -1 1) 0))))
(define yz-mirror (mirror 0))		;flip about yz-plane
(define zx-mirror (mirror 1))
(define xy-mirror (mirror 2))

(define (dilatation s)			;scale
  (matrix (vector s 0 0 0)
          (vector 0 s 0 0)
          (vector 0 0 s 0)
          (vector 0 0 0 1)))

; Perspective transformation.
; f = focal length = [retina radius]/(asin theta) where 2*theta = angle of view

(define (axonometric w f)
  (matrix (vector 1 0 0       0)
	  (vector 0 1 0       0)
	  (vector 0 0 1       0)    ;(/ 1 w)
	  (vector 0 0 (/ 1 f) 1)))  ;(/ 1 w) (/ f w)

; Compute a rotation matrix that will move a vector v onto the z-axis,
; and also, if possible, take an "up" vector to a position on the
; positive yz-plane.
;
; 1. Rotate about z axis to move v onto the xz-plane.
; 2. Rotate about y axis to move v onto the z-axis.
; 3. Rotate about z axis so that UP lies in the positive yz-plane.

(define (moving-to-z-axis v up)
  (let* ((x (x-component v))
	 (y (y-component v))
	 (z (z-component v))
	 (m (if (and (= x 0) (= y 0))
		(identity 4)
		(let ((theta (- (atan y x)))
		      (phi (- (atan (sqrt (+ (square x) (square y)))
				    z))))
		  ;; (format #t "~& xy: ~,2f  zx: ~,2f~%" theta phi)
		  (mm* (y-rotation phi)
		       (z-rotation theta)))))
	 (test (mv* m up)))
    (let ((x (x-component test))
	  (y (y-component test)))
      (if (and (= x 0) (= y 0))
	  m
	  (mm* (z-rotation (atan x y))
	       m)))))

; Views and their application.

(define (view-from-position camera position heading up)
  (let* ((f (camera-focal-length camera))
	 (retina (h+ position (h* (unitize heading) f))))
    (list (mm* (camera-transform camera)
	       (moving-to-z-axis (mv* (translation (h* retina -1)) position)
				 up)
	       ;; Put viewing plane at origin
	       (translation (h* retina -1)))
	  position
	  retina
	  up)))

(define view-transform car)
(define view-position cadr)
(define view-retina caddr)

(define (apply-view view v)
  (windowize (mv* (view-transform view) v)))

(define (windowize v)
  (let ((h (h-component v)))
    (list (round (/ (x-component v) h))
	  (round (/ (y-component v) h)))))

; Cameras.
; w = width of image area (retina)
; f = focal length

(define (make-camera w f width height lossage)
  (list (mm* lossage
	     (dilatation (/ (min width height) (* w 2)))
	     (axonometric w f)
	     xy-mirror)
	f w width height))

(define camera-transform car)
(define camera-focal-length cadr)
(define camera-retina-width caddr)
(define camera-width cadddr)
(define (camera-height c) (cadddr (cdr c)))
