
; Examples of use of geometry.scm package.

; To load and initialize:
;  (benchmark-mode)
;  (load "geometry")
;  (lisp:load "win")
;  (load "graphics")
;  (initialize-window)
;  (load "hax")

; This is no longer needed because initialize-window does it.
; (change-pen :line-width 2)  ;Thicker lines.

(define d 100)

(define view #f)

(define (set-position! new)
  (set! view (view-from-position *camera* new (h* new -1) (vector 0 0 1 1))))
(set-position! (vector (* d .2) (* d -2) (* d .8) 1))

(add-sync-action! (lambda () 
		    (set-position! (view-position view))))  ;Ugh.


; Platonic solids

(define tetra
  (copy (make-tetrahedron (floor (* d .8))) yz-mirror))
(define cube (make-cube (floor (* d .8))))
(define octa (dual (make-cube (floor (* d 1.3)))))
(define icosa (make-icosahedron (floor (* d .7))))
(define dodeca (dual (make-icosahedron (floor (* d .9)))))
(define platonic-solids (list tetra cube octa dodeca icosa))

(define cubocta (full-truncate cube))  ;= (full-truncate octa)
(define icosidodeca (full-truncate icosa))  ;= (full-truncate dodeca)


; Demonstrate some fun things.

(define (demo)
  (call-with-current-continuation
    (lambda (abort)
      (set! *abort* abort)
      (clear)(showit)
      (demo1 (let ((g (n-gon 6 (* d .6)) ))
	       (prism g (edge-length g) (face-normal g)))
	     #t)
      (for-each (lambda (s) (demo1 s #t)) platonic-solids)
      (demo1 cubocta #f)
      (demo1 icosidodeca #f)
      (pause " ")
      (shade (semi-truncate icosidodeca) view)
      (pause " ")
      (shade (semi-truncate dodeca) view)))
  (values))
(define *abort* (lambda (x) 'abort))

(define (demo1 solid dualize?)
  (pause (length (boundary solid)))
  (slick solid view)
  (pause "partial truncation")
  (operate semi-truncate solid)
  (pause "full truncation")
  (operate full-truncate solid)
  (if dualize?
      (begin (pause "dual")
	     (operate dual solid))))

(define (operate op x)
  (let ((d (op x)))
    (shade d view)
    (hide1 x view)
    (showit)))

(define (pause prompt)
  (newline)
  (display prompt)
  (if (string-ci=? (read-line (current-input-port)) "q")
      (*abort* #f)))

(define (read-line port)
  (let loop ((l '()))
    (let ((c (read-char port)))
      (if (eof-object? c)
	  c
	  (if (char=? c #\newline)
	      (list->string (reverse l))
	      (loop (cons c l)))))))

; Truncated polyhedra
;(define trunc-tetra (semi-truncate tetra))
;(define trunc-cube (semi-truncate cube))
;(define trunc-octa (semi-truncate octa))
;(define trunc-dodeca (semi-truncate dodeca))
;(define trunc-icosa (semi-truncate icosa))    ;soccer ball

; (full-truncate cubocta) ;is the small rhombicuboctahedron, almost 
; (semi-truncate cubocta) ;is the great rhombicuboctahedron, almost

; (full-truncate icosidodeca) ;is the small rhombicosidodecahedron, almost
; (semi-truncate icosidodeca) ;is the great rhombicosidodecahedron, almost

; (semi-truncate trunc-octa)
; (semi-truncate trunc-icosa)
; (semi-truncate trunc-cube)
; (full-truncate trunc-cube)
; etc. etc.

; The reason that (semi-truncate icosidodeca) is only *almost* the great
; rhomicisidodecahedron is that the edge lengths aren't all the same, so
; the faces aren't all regular polygons.  The topology and, I think, the
; symmetry group should be correct, though.



; Four-dimensional polyhedra (polytopes)

(define hypercube
  (let ((cube (make-cube (floor (* d .6)))))
    (prism cube (edge-length cube) (vector 4 0 3 1/4))))

(define hyperpyramid
  (cone (make-tetrahedron (* d .8))
	(make-vertex (vector (* d .55) 0 (* d .05) 1))))

; show-solids is useful for examining polytopes.  It displays one
; solid (e.g. one cube of a hypercube) at a time, pausing for a second
; after each one.

(define (show-solids cell)
  (frame cell view)
  (for-each-solid (lambda (s)
		   (format #t "~s " (unique-id s))
		   (with-color color/red
		     (lambda ()
		       (frame1 s view)))
		   (showit)
		   (sleep one-second)
		   (frame1 s view))
		 cell)
  (showit))


; I forget what this does.  Iterative truncation with line drawings, I guess.

(define (frob shape colors)
  (clear)
  (do ((colors colors (cdr colors))
       (shape shape (full-truncate shape)))
      ((null? colors) shape)
    (with-color (car colors) ;up to 10
      (lambda ()
	(hide1 shape view)
	(showit)))))

; Iterative truncation

(define (gt obj inc)
  (slick obj view)
  (do ((i (- 1 inc) (- i inc)))
      ((<= i 0)
       (pause 0)
       (slick (full-truncate obj) view))
    (pause i)
    (slick (general-truncate obj i) view)))


; I wrote this to help debug the face orientation logic.

(define (show-face-normal face view)
  (let* ((dface (dilate face 0.8))
	 (es (ordered-edges dface)))
    (with-color color/red  (lambda () (draw-edge (car es) view)))
    (with-color color/blue (lambda () (draw-edge (cadr es) view)))
    (for-each (lambda (e) (draw-edge e view))
	      (cddr es))
    (let ((c (centroid face))
	  (n (face-normal face)))
      (show-segment c (h+ c (h/ n (/ (hnorm n) *show-normal-length*)))
		    view))
    (let ((p (apply-view view (centroid face))))
      (scrawl (car p)
	      (cadr p)
	      (number->string (unique-id face))))
    (showit)))

(define (show-normals cell view)
  (clear)
  (with-line-width 1
    (lambda ()
      (frame1 cell view)))
  (with-line-width 2
    (lambda ()
      (for-each-visible-face
       (lambda (face)
	 (show-face-normal face view))
       cell view)))
  (values))

(define *show-normal-length* (floor (/ d 3)))

(define (show-segment v1 v2 view)
  (frame1 (make-edge (make-vertex v1)
		     (make-vertex v2))
	  view))


; Call SHOW-AXES if you're disoriented.

(define (show-axes w view)
  (let* ((o (make-vertex (vector 0 0 0 1)))
	 (show-axis (lambda (color v string)
		      (let ((z (make-vertex v)))
			(with-color color
			  (lambda ()
			    (draw-edge (make-edge o z) view)))
			(annotate z string view)))))
    (show-axis color/blue (vector w 0 0 1) "x")
    (show-axis color/red (vector 0 w 0 1) "y")
    (show-axis color/green (vector 0 0 w 1) "z")
    (showit)))
    
(define (show-xy width view)		;e.g. (show-xy (* d .6) view)
  (let ((v1 (make-vertex (vector width width 0 1)))
	(v2 (make-vertex (vector (- width) width 0 1)))
	(v3 (make-vertex (vector (- width) (- width) 0 1)))
	(v4 (make-vertex (vector width (- width) 0 1))))
    (with-color color/green
      (lambda ()
	(frame1 (make-cell (list (make-edge v1 v2)
				 (make-edge v2 v3)
				 (make-edge v3 v4)
				 (make-edge v4 v1)))
		view)))
    (showit)))

(define (annotate-faces cell proc)
  (clear)
  (for-each-visible-face
     (lambda (face)
       (frame1 face view)
       (annotate face (proc face) view))
     cell view)
  (showit))

; Try (foo icosa)

(define (foo x)
  (shade x view)
  (pause "--<return> for more--")
  (let ((y (semi-truncate x)))
    (shade y view)
    (pause "--<return> for more--")
    (foo (dual y))))


;; Products
;
;; Try these:
;;   (show-solids hypercube)
;;   (hide* torus view)
;;   (frame (product (dilate tetra .1) (dilate cube .8) h+) view)
;;   etc.
;
;; Hypercube
;
;(define e1 (make-edge (make-vertex (vector -30 0 0 1))
;                      (make-vertex (vector 30 0 0 1))))
;(define e2 (copy e1 (z-rotation (/ pi 2))))
;(define e3 (copy e1 (y-rotation (/ pi 2))))
;(define e4 (copy e1 (mm* (z-rotation (/ pi 4))
;                         (y-rotation (/ pi -4)))))
;
;(define hypercube (product (product e1 e2 h+) (product e3 e4 h+) h+))
;
;; Torus
;
;; The torus doesn't have the usual torus geometry.  That would require
;; something more sophisticated than vector addition for computing the
;; coordinates of the product of two vertices.
;
;(define g1 (boundary (n-gon 13 40)))
;(define g2 (copy g1 (mm* (dilatation 0.5) (y-rotation (/ pi 2)))))
;
;(define torus (product g1 g2 h+))
;
;; Also try (hide* (product g1 e2 h+) view), etc.
